//---------------------------------------------------------------------------
// Copyright (c) 1999,2000 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//-------------------------------------------------------------------------
//
//  Description:
//      This is a class that extends JTable. The device and property is set 
//      in synchronous. It is possible to insert, replace or delete any of
//      the devices or properties by readjusting the table to conform to  
//      the needs as the user wishes.
//
//  Author: Masa Ebisawa, Ying Chen
//
//  Based on work by Chip Watson
//                                       
//-------------------------------------------------------------------------
// Revision History:
//  05/??/00    Release 0.0  Masa Ebisawa 
//  06/29/00    Release 0.1  Reorganize the code (yc) 
//
//--------------------------------------------------------------------------

package sdm;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import cdev.*;
import cdev.data.*;

/**
 * This class extends JTable to include some features useful for displaying 
 * multiple information in an organized manner. We added features so that
 * the table can insert, replace and delete any of the rows and columns as
 * needed. Also this class implements TableColumnListener, MouseListener and
 * ListSelectionListener to know when thing are modified.
 */
public class CTable extends JTable implements TableColumnModelListener,
					      MouseListener,
					      Scrollable, 
                                              ActionListener { 

    static final String BLANK = "";
    protected cdev.Device[] device;
    protected boolean active = false, changed = false;
    protected String[] deviceNames, propertyNames, temp;
    protected String[] lastDeviceNames, lastPropertyNames;
    protected cellListener[][] cellListeners;
    protected int maxRow, maxCol, pos = 0, maxWidth = 5;
    protected DefaultTableModel myModel;
    protected DefaultTableColumnModel myColModel;
    protected Vector columnWidth;
    protected JPopupMenu pop;
    protected JMenuItem insert, replace, delete; 
    private char type_id = '0', rowORcol = '0';


    /**
     * Create an empty table
     */
    public CTable(){
	super();
    }

    /**
     * Creates a table with the given device names and property names.
     * @param rownames    array of Devices.
     * @param colnames    array of properties.
     */
    public CTable(String[] rownames, String[] colnames){ 
	super();
	create(rownames, colnames);
	columnWidth = new Vector();
	for(int c=0; c<maxCol; c++)
	    columnWidth.add(new Integer(myColModel.getColumn(c).getWidth()));
    }
    
    /**
     * Creates a table with the given device names and property names
     * and also the column width.
     * @param rownames    This is an array of Devices.
     * @param colnames    This is an array of properties.
     * @param colWidth    This is an array of column Widths.
     */
    public CTable(String[] rownames, String[] colnames,
		      int[] colWidth){
	super();
	columnWidth = new Vector();
	columnWidth.add(new Integer(20));
	for(int i=0; i<colWidth.length; i++){
	    columnWidth.add(new Integer(colWidth[i]));
	}
	create(rownames, colnames);
    }

    /**
     *function for the constructor to create the initial table;
     */
    private void create(String[] rownames, String[] colnames){
	String[] temp = fixColumnNames(colnames);
	this.setModel(new DefaultTableModel(temp, rownames.length));
	myColModel = (DefaultTableColumnModel)this.getColumnModel();
	myColModel.setColumnSelectionAllowed(true);
	DefaultTableCellRenderer tempRender = new DefaultTableCellRenderer();
	tempRender.setHorizontalAlignment(JLabel.CENTER);	

	maxRow = rownames.length;
	maxCol = temp.length;
	setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
	cellListeners = new cellListener[maxRow][maxCol];	
	for(int c = 1; c<maxCol; c++) {
	    for(int r = 0; r<maxRow; r++)
		cellListeners[r][c] = new cellListener(r, c);
	}	    
	for(int c = 1; c<maxCol; c++) {
	    myColModel.getColumn(c).setCellRenderer(tempRender);
	}
	setDeviceNames(rownames);
	setPropertyNames(temp);
	getTableHeader().addMouseListener(new headerMouseListener());	
	addMouseListener(this);
    }


    /**
     * method that fires when the table has been changed
     */
     protected void propChange() {
	 firePropertyChange("", null, null);
    }

    /**
     * Just in case the table doesn't have the right amount of columns
     * especially to set the device names this is called just in case. 
     * @param  properties    The array of properties passed in to the CTOR.
     */
     private String[] fixColumnNames(String[] properties){
	if (properties[0].equals(" ")) 
	    return properties;
	else {
	    String[] columnNames = new String[properties.length+1];
	    for(int i=0; i<properties.length+1; i++) {
		if(i==0)  columnNames[i] = " ";
		else      columnNames[i] = properties[i-1];
	    }
	    return columnNames;
	}
     }
    

    private void setupCellListener () {
	cellListener[][] tempCellListener = new cellListener[maxRow][maxCol];
	for(int c=1; c<maxCol; c++){
	    for(int r = 0; r<maxRow; r++)
		tempCellListener[r][c] = new cellListener(r, c);
	}
	cellListeners = tempCellListener;	
    }

    /**
     * This method is called synchronous to setDeviceName at row r.
     *
     * @param r          row index for new device
     * @param newName    new device names to be set.
     */
    synchronized public void setDeviceName (int r, String newName) {
	System.out.println("setDeviceName ("+newName+") at "+r);
	newName = CUtility.makeValidName(newName);
	if (!CUtility.testNameMatch(deviceNames[r], newName)) {
	    try {
		for(int c=1; c<maxCol; c++) 
		    stopMonitor(r, c);
		deviceNames[r] = newName;
		lastDeviceNames[r] = deviceNames[r]; 
		device[r] = new Device(deviceNames[r]); 
		changed = true;
		setDeviceHeader();
		for(int c=1; c<maxCol; c++) 
		    startMonitor(r, c);
	    } catch(Exception e){ e.printStackTrace(); }
	}
    }
							      
    /**
     * This method is called synchronous to set deviceNames. 
     * 
     * @param newNames   The array of device names that are passed.
     */
    synchronized public void setDeviceNames(String[] newNames){
	try {
	    for(int c=1; c<maxCol; c++) {
		for(int r = 0; r<maxRow; r++) 
		    stopMonitor(r, c);
	    } 
	    maxRow = newNames.length;
	    if(deviceNames == null || deviceNames.length != newNames.length){
		device = new Device[newNames.length];
		deviceNames = new String[newNames.length];
		lastDeviceNames = new String[newNames.length];
	    }
	    if (propertyNames != null) setupCellListener();
	    for(int r = 0; r<maxRow; r++) {
		newNames[r] = CUtility.makeValidName(newNames[r]);
		if (!CUtility.testNameMatch(deviceNames[r], newNames[r])) {	 
		    deviceNames[r] = newNames[r];
		    lastDeviceNames[r] = deviceNames[r]; 
		    device[r] = new Device(deviceNames[r]);
		}
	    }
	    setDeviceHeader();
	    changed = true;
	    for (int c=1; c<maxCol; c++) {
		for(int r = 0; r<maxRow; r++)
		    startMonitor(r, c);
	    }
	} catch(Exception e) { e.printStackTrace(); }
    }

    /**
     * Returns the array of device names.
     */
    public String[] getDeviceNames() {
	String[] copy = (String[])deviceNames.clone();
	return copy;
    }

    /**
     *  Set propertyName at column c.
     *
     * @param c          column index for new property
     * @param newName    new property names to be set.
     */
    synchronized public void setPropertyName (int c, String newName){
	newName = CUtility.makeValidName(newName);
	if (!CUtility.testNameMatch(propertyNames[c], newName)) {
	    try {
		for(int r=0; r<maxRow; r++) 
		    stopMonitor(r, c+1);
		propertyNames[c+1] = newName;
		lastPropertyNames[c+1] = propertyNames[c+1]; 
		setDataModel();
		setDeviceHeader();
		changed = true;
		for(int r=0; r<maxRow; r++) 
		    startMonitor(r, c+1);
	    } catch(Exception e){ e.printStackTrace(); }
	}
    }

    /**
     * This method is called synchronous to setPropertyNames. 
     * 
     * @param newNames   The array of property names that are passed.
     */
    synchronized public void setPropertyNames(String[] newNames){
	try{
	    for (int c=1; c<maxCol; c++) {
		for(int r = 0; r<maxRow; r++) 
		    stopMonitor(r, c);
	    }
	    maxCol = newNames.length;
	    if(propertyNames==null || propertyNames.length !=newNames.length) {
		propertyNames = new String[newNames.length];
		lastPropertyNames = new String[newNames.length];
	    }
	    if (deviceNames != null)  setupCellListener();
	    for(int c=0; c<maxCol; c++){
		newNames[c] = CUtility.makeValidName(newNames[c]);
		if (!CUtility.testNameMatch(propertyNames[c], newNames[c])) {	
		    propertyNames[c] = newNames[c];
		    lastPropertyNames[c] = propertyNames[c];
		}
	    }
	    setDataModel();
	    setDeviceHeader(); 
	    changed = true;
	    for(int c=1; c<maxCol; c++) {
		for(int r = 0; r<maxRow; r++)
		    startMonitor(r, c);
	    } 
	} catch(Exception e){ e.printStackTrace(); }
    }
    
    /**
     * Returns the array of property names.
     */
    public String[] getPropertyNames(){
	String[] names = new String[maxCol-1];
	for(int i=1; i<maxCol; i++)
	    names[i-1] = propertyNames[i];
	return names;
    }

    /**
     * Returns the size of the column at index i 
     * (not include the row head column)
     * @param    the index of column.
     */
    public int getColumnWidth (int i) {
	return ((Integer)columnWidth.elementAt(i+1)).intValue();
    }
    
    /**
     * Returns the array of column widths to the utility.
     * @return array of column widths
     */
    public int[] getColumnWidths() {
	int[] val = new int[maxCol-1];
	for(int i=1; i<maxCol; i++)
	    val[i-1] = ((Integer)columnWidth.elementAt(i)).intValue();
	return val;
    }

    /**
     * to be implemented next release
     */
    public void setColumnWidths (int[] width) { }

    /**
     * to be implemented next release
     * sets the column width of the selected column index 
     * @param index        index of the column
     * @param size         size of the column
     */
    public void setColumnWidth (int index, int size) { }

    /**
     * stops the monitor on a specific cell
     * @param r   The row index
     * @param c   The column index
     */
    protected void stopMonitor(int r, int c) {
	if (cellListeners[r][c].active) {
	    try {	
		cellListeners[r][c].active = false;
		device[r].monitorOff(propertyNames[c], cellListeners[r][c]);
	    } catch (IOException e) { e.printStackTrace(); }
	}
    }

    /**
     * starts the monitor on a specific cell
     * param r   The row index
     * param c   The column index
     */
    protected void startMonitor(int r, int c) {
	try {
	    if (cellListeners[r][c].active) stopMonitor(r, c);
	    if (deviceNames == null || propertyNames == null) return;
	    if (deviceNames[r] == null || propertyNames[c] == null) return;
	    if (!deviceNames[r].equals(lastDeviceNames[r])) {
		device[r] = new Device(deviceNames[r]);
		lastDeviceNames[r] = deviceNames[r];
	    }
	    device[r].monitorOn(propertyNames[c], cellListeners[r][c]);
	    lastPropertyNames[c] = propertyNames[c];
	    cellListeners[r][c].active = true;
	} catch (IOException e) {
	    cellListeners[r][c].active = false;
	    e.printStackTrace();
	}
    }

    
    /**
     * This sets the data model whenever necessary.
     */
    private void setDataModel(){
	this.setModel(new DefaultTableModel(propertyNames, maxRow));
	myModel = (DefaultTableModel)this.getModel();
	if (columnWidth != null) {
	    int val;
	    for (int c=0; c<maxCol; c++){
		val = ((Integer)(columnWidth.elementAt(c))).intValue();
		myColModel.getColumn(c).setPreferredWidth(val);
		myColModel.getColumn(c).setWidth(val);
	    }
	}
    }
    
    /**
     * This sets the headers for the devices whenever it needs updating
     */
    private void setDeviceHeader() {
	int maxSize = 0;
	String tempString = new String();
	for(int i=0; i<maxRow; i++) {
	    if(deviceNames[i].length() > maxSize) {
		maxSize = deviceNames[i].length();
		tempString = deviceNames[i];
	    }
	}
	FontMetrics fm = getAccessibleContext().getAccessibleComponent()
	                 .getFontMetrics(this.getFont());
	maxWidth = fm.stringWidth(tempString) + (maxSize*2) + 5;
	DefaultTableCellRenderer CR = new DefaultTableCellRenderer();
	//CR.setBackground(Color.lightGray);
	myColModel.getColumn(0).setCellRenderer(CR);
	myColModel.getColumn(0).setPreferredWidth(maxWidth);
	myColModel.getColumn(0).setMinWidth(maxWidth);
	myColModel.getColumn(0).setMaxWidth(maxWidth);
	myColModel.getColumn(0).setWidth(maxWidth);
	for(int r=0; r<maxRow; r++)
	    this.setValueAt(deviceNames[r], r, 0);
	propChange();
    }
    
    /**
     * Return true if table state change otherwise return false.
     */
    public boolean getTableState() {
	// It used by the utility program to see if there has been 
	// anyc hanges made to the table.
       	boolean temp = changed;
    	changed = false;
    	return temp;
    }

    /**
     * Remove device at given index
     * @param position   device index to be removed
     */
    public void removeDevice (int position) {
	myModel.removeRow(position);  
	int max = maxRow - 1;
	String[] newdeviceNames = new String[max];
	for(int r = 0; r<max; r++) {
	    if (r<position)
		newdeviceNames[r] = deviceNames[r];
	    else 
		newdeviceNames[r] = deviceNames[r+1]; 
	}
	setDeviceNames(newdeviceNames);	 
	setDeviceHeader();
	propChange();
    }

    /**
     * Remove property at given index
     * @param position   property index to be removed
     */
    public void removeProperty (int position) {
	int max = maxCol - 1;
	String[] newPropertyNames = new String[max];
	for(int c=0; c<max; c++) {
	    if (c < position+1) 
		newPropertyNames[c] = propertyNames[c];
	    else
		newPropertyNames[c] = propertyNames[c+1];
	}
	setPropertyNames (newPropertyNames);
	columnWidth.removeElementAt (position+1);
	validate();
    }
    
    /**
     * Insert device at given index
     * @param position   device index to be inserted
     * @param newName    new device name to be inserted
     */
    public void insertDevice (int position, String newName) {
	int max = maxRow + 1;
	if(position == max)
	    myModel.addRow(propertyNames);
	else
	    myModel.insertRow(position, propertyNames);

	String[] newdeviceNames = new String[max];
	for (int r = 0; r<max; r++) {
	    if (r < position) 
		newdeviceNames[r] = deviceNames[r];
	    else if (r > position) 
		newdeviceNames[r] = deviceNames[r-1]; 
	    else
		newdeviceNames[position] = newName;	    
	}
	setDeviceNames(newdeviceNames);
	propChange();
    }

    /**
     * Insert property at given index
     * @param position   property index to be inserted
     * @param newName    new property name to be inserted
     */
    public void insertProperty (int position, String newName) {
	int max = maxCol+1;
	String[] newPropertyNames = new String[max];
	for (int c = 0; c<max; c++) {
		if(c == position + 1)
		    newPropertyNames[c] = newName;
		else if(c < position + 1)
		    newPropertyNames[c] = propertyNames[c];
		else
		    newPropertyNames[c] = propertyNames[c-1];	    
	    }
	columnWidth.insertElementAt(new Integer(75), position+1);
	setPropertyNames(newPropertyNames);
	validate();
    }

    /**
     * handle the popup menu insert, replace and delete
     */
    public void actionPerformed (ActionEvent event) {
	if (event.getSource() == insert) {
	    type_id = 'I';
	    (new RequestInfo (this, type_id, rowORcol, pos)).show();
	}
	else if (event.getSource() == replace) {
	    type_id = 'R';
	    (new RequestInfo (this, type_id, rowORcol, pos)).show();
	}
	else if (event.getSource() == delete) {
	    type_id = 'D';
	    if (rowORcol == 'C')     removeProperty (pos);
	    else if (rowORcol == 'R')  removeDevice (pos);
	}
    }

    private void createPopup () {
	if (pop == null) {
	    pop = new JPopupMenu();
	    insert = new JMenuItem("Insert"); 
	    replace = new JMenuItem("Replace"); 
	    delete = new JMenuItem("Delete");
	    pop.add(insert);
	    pop.add(replace);
	    pop.add(delete);
	    insert.addActionListener(this);
	    replace.addActionListener(this);
	    delete.addActionListener(this);
	}
    }

    /**
     * popup a menu when right mouse down
     * @param e  A MouseEvent 
     */    
    public void mousePressed(MouseEvent e) {
	if ( e.getX()/maxWidth == 0) {
	    pos = e.getY()/(getRowHeight() + 1);
	    //System.out.println("press on row "+pos);
	    rowORcol = 'R';
	    createPopup();
	    if (e.isPopupTrigger()) 
		pop.show(this, e.getX(), e.getY());
	}
    }

    public void mouseReleased(MouseEvent e) {
	if (e.isPopupTrigger()) 
	    pop.show(this, e.getX(), e.getY());
    }

    public void mouseClicked (MouseEvent e) { }
    public void mouseEntered (MouseEvent e) { }
    public void mouseExited  (MouseEvent e) { }

 
 //*********************************************************************

    /**
     * This class implements deviceListener which has a fuction
     * deviceChanged(). This gets the data from the system and puts it 
     *into the correct cell.
     */    
    protected class cellListener implements DeviceListener{
	int row;
	int col;
	boolean active = false;
	
	/**
	 * constructor
	 * @param r    row index
	 * @param c    column index
	 */
	cellListener(int r, int c) {
	    row = r;
	    col = c;
	}
	
	/**
	 * method inhereted from DeviceListener
	 * @param event  A DeviceEvent
	 */
	public void deviceChanged(DeviceEvent de) {
	    try {
		String val=de.getData().getDataEntry("value").stringValue();
		myModel.setValueAt(val, row , col);	
	    } catch (Exception e) { e.printStackTrace(); }
	}
    }


 //************************************************************************

    /**
     * Inner class that implements MouseListener to help detect mouse 
     * events that occur on the column headers.
     */
    protected class headerMouseListener implements MouseListener {

	protected TableColumn tabcol;
	protected int viewColumn, resizeColumn;

	public void mousePressed(MouseEvent e) {
	    resizeColumn = 0;
	    viewColumn = myColModel.getColumnIndexAtX(e.getX());
	    JTableHeader t = getTableHeader();
	    if(t.getResizingColumn() != null) {
		tabcol = t.getResizingColumn();
		resizeColumn = tabcol.getModelIndex();
	    }
	    if (viewColumn != 0) {
		rowORcol = 'C';
		pos = viewColumn - 1;
		createPopup();
		if (e.isPopupTrigger())
		    pop.show(e.getComponent(), e.getX(), e.getY());
	    }
	}

	public void mouseReleased(MouseEvent e) {
	    if ((viewColumn != 0) && e.isPopupTrigger())
		pop.show(e.getComponent(), e.getX(), e.getY()); 
	    if(resizeColumn != 0) {
		Integer w = new Integer(tabcol.getWidth());
		columnWidth.setElementAt(w, resizeColumn);
		propChange();
		tabcol = null;
	    }
	}

	public void mouseClicked(MouseEvent e) { }
	public void mouseEntered(MouseEvent e) { }
	public void mouseExited (MouseEvent e) { }

    }
}


//**************************************************************************

/**
 * A class that extends JDialog an is used for inserting, 
 * deleting, replacing devices and properties. This frame gives you 
 * the option of what to insert and either to insert beforee or after 
 * the specified location.
 */
class RequestInfo extends JDialog implements ActionListener {  

    public CTable table;
    private char type, row_col;
    private int position;
    JComboBox input;
    JButton okB, cancelB;
    JRadioButton before, after;

    public RequestInfo (CTable parent, char id, char r_c, int i) {
	table = parent;
	type = id;
	row_col = r_c;
	position = i;
	createDialog();
    }

    private void createDialog () {
	String title = "", instruction = "";
	if (type == 'I')            title = "Insert ";
	else if (type == 'R')       title = "Replace ";
	if (row_col == 'R') {
	    title = title+"Row";
	    instruction = "Device Name:";
	}
	else if (row_col == 'C') {
	    title = title+"Column";
	    instruction = "Property Name:";
	}
	setTitle(title);

	JPanel centerP = new JPanel();
	centerP.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 10));
	input = new JComboBox(table.getDeviceNames());
	input.setMaximumSize(new Dimension(4,1));
	input.setEditable(true);
	DefaultComboBoxModel cBoxModel = new DefaultComboBoxModel();
	input.setModel(cBoxModel);
	if (type == 'R') {
	    if (row_col == 'R')
		input.addItem(table.getDeviceNames()[position]);
	    else if (row_col == 'C')
		input.addItem(table.getPropertyNames()[position]);
	}
	JPanel p1 = new JPanel();
	p1.setLayout(new GridLayout(2, 1, 20, 5));
	p1.add(new JLabel(instruction), BorderLayout.NORTH);
	p1.add(input, BorderLayout.CENTER);
	centerP.add(p1);


	if (type == 'I') {
	    JPanel p2 = new JPanel();
	    p2.setLayout(new GridLayout(2, 1, 20, 5));
	    before = new JRadioButton("Before", true);
	    after  = new JRadioButton("After ", false);
	    ButtonGroup bg = new ButtonGroup();
	    bg.add(before);
	    bg.add(after);
	    p2.add(before);
	    p2.add(after);
	    centerP.add(p2);
	}

	JPanel southP = new JPanel();
	southP.setLayout(new FlowLayout(FlowLayout.CENTER, 10, 10));
	okB = new JButton("OK"); 
	okB.addActionListener(this);
	okB.setMnemonic(KeyEvent.VK_ENTER);
	cancelB = new JButton("Cancel");
	cancelB.addActionListener(this);
	southP.add(okB);
	southP.add(cancelB);

	Container content = getContentPane();
	content.add(centerP, BorderLayout.CENTER);
	content.add(southP, BorderLayout.SOUTH);
	pack();
	this.requestFocus();
    }
    
    public void actionPerformed(ActionEvent event){
	String msg = "";  
	if(event.getSource() == cancelB) 
	    setVisible(false);
	else if (event.getSource() == okB) {
	    if (input.getSelectedItem() != null) {
		performWork();
		setVisible(false);
	    }
	    else {
		if (row_col == 'C')   msg="You must provide Property Name!";
		else if(row_col=='R') msg="You must provide Device Name!";
		JOptionPane.showMessageDialog(this, msg, "Warning!!!", 
					      JOptionPane.WARNING_MESSAGE);
	    }
	}
    }

    private void performWork() {
	String newName = (String) input.getSelectedItem();
	if (type == 'R') {
	    if (row_col == 'C')   table.setPropertyName (position, newName);
	    else if (row_col=='R')  table.setDeviceName (position, newName);
	}
	else if (type == 'I') {
	    if (after.isSelected())  position ++;
	    if (row_col == 'C')      table.insertProperty(position, newName);
	    else if (row_col=='R')   table.insertDevice  (position, newName);
	}
    }


}



